""" Contains all game parameters in one place """

from colorama import init, Fore, Style, Back
init(autoreset=True)


class Lives:
    lives = 3


# Define scene length(vertical), width(horizontal),
# fullwidth(map length)
SC_LENGTH = 40
SC_WIDTH = 100
SC_FULLWIDTH = 500

MARIO_LENGTH = 4
MARIO_WIDTH = 3
MARIO_INIT_X = 32
MARIO_INIT_Y = 4

SCENE_MOVE_AFTER = 60

ALLOWED_INPUTS = ['w', 'a', 'd', 'A', 'C', 'D', ' ']

SCORE_DIST = 10
SCORE_KILLED = 500
SCORE_COINS = 200

SPEED = {1: 0.25,
         2: 0.2,
         3: 0.15}
# color escape sequences
colors = {
    'Blue': '\x1b[0;34m',
    'Green': '\x1b[0;32m',
    'Cyan': '\x1b[0;36m',
    'Red': '\x1b[0;31m',
    'Purple': '\x1b[0;35m',
    'Brown': '\x1b[0;33m',
    'Gray': '\x1b[0;37m',
    'Light Green': '\x1b[1;32m',
    'Light Cyan': '\x1b[1;36m',
    'Yellow': '\x1b[1;33m',
    'White': '\x1b[1;37m'
}
RESET = '\x1b[0m'

groundx = 36
barriers = [colors['Brown'] + '#' + RESET, '#']
killers = [colors['Blue'] + '~' + RESET, '~', '{', '}']
beaters = ['^']
monies = [colors['Yellow'] + '$' + RESET]



# Stubs for statistics

from decimal import Decimal
from fractions import Fraction
import sys
from typing import Iterable, Optional, TypeVar

# Most functions in this module accept homogeneous collections of one of these types
_Number = TypeVar('_Number', float, Decimal, Fraction)

class StatisticsError(ValueError): ...

def mean(data: Iterable[_Number]) -> _Number: ...
if sys.version_info >= (3, 6):
    def harmonic_mean(data: Iterable[_Number]) -> _Number: ...
def median(data: Iterable[_Number]) -> _Number: ...
def median_low(data: Iterable[_Number]) -> _Number: ...
def median_high(data: Iterable[_Number]) -> _Number: ...
def median_grouped(data: Iterable[_Number]) -> _Number: ...
def mode(data: Iterable[_Number]) -> _Number: ...
def pstdev(data: Iterable[_Number], mu: Optional[_Number] = ...) -> _Number: ...
def pvariance(data: Iterable[_Number], mu: Optional[_Number] = ...) -> _Number: ...
def stdev(data: Iterable[_Number], xbar: Optional[_Number] = ...) -> _Number: ...
def variance(data: Iterable[_Number], xbar: Optional[_Number] = ...) -> _Number: ...



#!/usr/bin/python3

# Copyright (c) 2014, Stephen Warren
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.import random

import random
import sys
import termios
import tty

DIR_LEFT = 0
DIR_RIGHT = 1
DIR_UP = 2
DIR_DOWN = 3

xs = ys = 4

from rp import *
def board_index(x, y):
    return (y * ys) + x

def print_board(board):
    for y in range(ys):
        print('+----' * xs + '+')
        for x in range(xs):
            val = board[board_index(x, y)]
            if val:
                cell = '% 4d' % val
                color='gray'
                colors={1:'red',2:'magenta',4:'cyan',8:'yellow',16:'blue',32:'gray',64:'green',128:'red',256:'magenta',512:'cyan',1024:'magenta',2048:'yellow'}
                if val in colors:
                    color=colors[val]
                cell=fansi(cell,color,'bold')
            else:
                cell = '    '
            print('|' + cell , end='')
        print('|')
    print('+----' * xs + '+')

def add_piece(board):
    empty_positions = [i for i, val in enumerate(board) if val == 0]
    insert_pos = random.randint(0, len(empty_positions) - 1)
    board[empty_positions[insert_pos]] = 1
    return board

def getch():
    fd = sys.stdin.fileno()
    orig_termios_attr = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, orig_termios_attr)
    return ch

rows_l2r_t2b = [[board_index(x, y) for x in range(xs)] for y in range(ys)]
rows_r2l_t2b = [[board_index(xs - x - 1, y) for x in range(xs)] for y in range(ys)]

cols_t2b_l2r = [[board_index(x, y) for y in range(ys)] for x in range(xs)]
cols_b2t_l2r = [[board_index(x, ys - y - 1) for y in range(ys)] for x in range(xs)]

pos_lists_of_move = {
    DIR_LEFT:  rows_l2r_t2b,
    DIR_RIGHT: rows_r2l_t2b,
    DIR_UP:    cols_t2b_l2r,
    DIR_DOWN:  cols_b2t_l2r,
}

def apply_move(board, move):
    pos_lists = pos_lists_of_move[move]
    score_delta = 0
    for pos_list in pos_lists:
        dst = 0
        for src in range(len(pos_list)):
            if board[pos_list[src]]:
                board[pos_list[dst]] = board[pos_list[src]]
                dst += 1
        for i in range(dst, len(pos_list)):
            board[pos_list[i]] = 0
        for i in range(len(pos_list[:-1])):
            if board[pos_list[i]] == board[pos_list[i + 1]]:
                board[pos_list[i]] += board[pos_list[i + 1]]
                score_delta += board[pos_list[i]]
                for j in range(i + 1, len(pos_list) - 1):
                    board[pos_list[j]] = board[pos_list[j + 1]]
                board[pos_list[-1]] = 0
    return board, score_delta

move_dir_map = {
    'h': DIR_LEFT,
    'j': DIR_DOWN,
    'k': DIR_UP,
    'l': DIR_RIGHT,
}

def get_apply_move(board):
    while True:
        print('Move: ', end='')
        sys.stdout.flush()
        ch = getch().lower()
        wasd={'a':'h','w':'k','s':'j','d':'l'}
        if ch in wasd:
            ch=wasd[ch]#turn wasd into vim keys
        print(ch)
        if ch == 'q':
            quit()
        if ch == 'p':
            print_board(board)
            continue
        if ch not in move_dir_map:
            print('Unknown keypress. Keys: movement:w/a/s/d/ or h/j/k/l , print: p, quit: q')
            continue
        move = move_dir_map[ch]
        new_board, score_delta = apply_move(board.copy(), move)
        if new_board == board:
            print('Illegal move')
            continue
        break
    print()
    return add_piece(new_board), score_delta

board = [0] * (ys * xs)
board = add_piece(board)
score = 0

detect_2048 = True
def quit():
    print('Quitting')
    high_score=0
    filename='.rp.2048.highscore'
    try:high_score=int(text_file_to_string(filename))
    except:pass
    message='Quit 2048: Your score was '+str(score)+'. The previous high score was '+str(high_score)
    high_score=max(high_score,score)
    try:string_to_text_file(filename,str(high_score))
    except:pass
    assert False,message
    #sys.exit(0)

while True:
    print_board(board)
    print('Score:', score)
    if detect_2048 and 2048 in board:
        while True:
            print('You win; Continue/Quit? ', end='')
            sys.stdout.flush()
            ch = getch().lower()
            print(ch)
            if ch == 'q':
                quit()
            if ch == 'c':
                print('Continuing')
                detect_2048 = False
                break
            print('Unknown option (c/q)')
    moves_left = False
    for move in (DIR_LEFT, DIR_RIGHT, DIR_UP, DIR_DOWN):
        new_board = apply_move(board.copy(), move)
        if new_board != board:
            moves_left = True
            break
    if not moves_left:
        print('You lose!')
        sys.exit(0)
    board, score_delta = get_apply_move(board)
    score += score_delta


from typing import Any
from .. import exceptions
from .. import packages

ConnectTimeoutError = exceptions.ConnectTimeoutError
MaxRetryError = exceptions.MaxRetryError
ProtocolError = exceptions.ProtocolError
ReadTimeoutError = exceptions.ReadTimeoutError
ResponseError = exceptions.ResponseError

log: Any

class Retry:
    DEFAULT_METHOD_WHITELIST: Any
    BACKOFF_MAX: Any
    total: Any
    connect: Any
    read: Any
    redirect: Any
    status_forcelist: Any
    method_whitelist: Any
    backoff_factor: Any
    raise_on_redirect: Any
    def __init__(self, total=..., connect=..., read=..., redirect=..., method_whitelist=..., status_forcelist=..., backoff_factor=..., raise_on_redirect=..., _observed_errors=...) -> None: ...
    def new(self, **kw): ...
    @classmethod
    def from_int(cls, retries, redirect=..., default=...): ...
    def get_backoff_time(self): ...
    def sleep(self): ...
    def is_forced_retry(self, method, status_code): ...
    def is_exhausted(self): ...
    def increment(self, method=..., url=..., response=..., error=..., _pool=..., _stacktrace=...): ...



from __future__ import unicode_literals

from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.contrib.completers import PathCompleter
from prompt_toolkit.contrib.regular_languages.compiler import compile as compile_grammar
from prompt_toolkit.contrib.regular_languages.completion import GrammarCompleter

from rp.rp_ptpython.utils import get_jedi_script_from_document

import re

__all__ = (
    'PythonCompleter',
)


class PythonCompleter(Completer):
    """
    Completer for Python code.
    """
    def __init__(self, get_globals, get_locals):
        super(PythonCompleter, self).__init__()

        self.get_globals = get_globals
        self.get_locals = get_locals

        self._path_completer_cache = None
        self._path_completer_grammar_cache = None

    @property
    def _path_completer(self):
        if self._path_completer_cache is None:
            self._path_completer_cache = GrammarCompleter(
                self._path_completer_grammar, {
                    'var1': PathCompleter(expanduser=True),
                    'var2': PathCompleter(expanduser=True),
                })
        return self._path_completer_cache

    @property
    def _path_completer_grammar(self):
        """
        Return the grammar for matching paths inside strings inside Python
        code.
        """
        # We make this lazy, because it delays startup time a little bit.
        # This way, the grammar is build during the first completion.
        if self._path_completer_grammar_cache is None:
            self._path_completer_grammar_cache = self._create_path_completer_grammar()
        return self._path_completer_grammar_cache

    def _create_path_completer_grammar(self):
        def unwrapper(text):
            return re.sub(r'\\(.)', r'\1', text)

        def single_quoted_wrapper(text):
            return text.replace('\\', '\\\\').replace("'", "\\'")

        def double_quoted_wrapper(text):
            return text.replace('\\', '\\\\').replace('"', '\\"')

        grammar = r"""
                # Text before the current string.
                (
                    [^'"#]                                  |  # Not quoted characters.
                    '''  ([^'\\]|'(?!')|''(?!')|\\.])*  ''' |  # Inside single quoted triple strings
                    "" " ([^"\\]|"(?!")|""(?!^)|\\.])* "" " |  # Inside double quoted triple strings

                    \#[^\n]*(\n|$)           |  # Comment.
                    "(?!"") ([^"\\]|\\.)*"   |  # Inside double quoted strings.
                    '(?!'') ([^'\\]|\\.)*'      # Inside single quoted strings.

                        # Warning: The negative lookahead in the above two
                        #          statements is important. If we drop that,
                        #          then the regex will try to interpret every
                        #          triple quoted string also as a single quoted
                        #          string, making this exponentially expensive to
                        #          execute!
                )*
                # The current string that we're completing.
                (
                    ' (?P<var1>([^\n'\\]|\\.)*) |  # Inside a single quoted string.
                    " (?P<var2>([^\n"\\]|\\.)*)    # Inside a double quoted string.
                )
        """

        return compile_grammar(
            grammar,
            escape_funcs={
                'var1': single_quoted_wrapper,
                'var2': double_quoted_wrapper,
            },
            unescape_funcs={
                'var1': unwrapper,
                'var2': unwrapper,
            })

    def _complete_path_while_typing(self, document):
        char_before_cursor = document.char_before_cursor
        return document.text and (
            char_before_cursor.isalnum() or char_before_cursor in '/.~')

    def _complete_python_while_typing(self, document):
        char_before_cursor = document.char_before_cursor
        return document.text and (
            char_before_cursor.isalnum() or char_before_cursor in '_.')

    def get_completions(self, document, complete_event):
        """
        Get Python completions.
        """
        # Do Path completions
        if complete_event.completion_requested or self._complete_path_while_typing(document):
            for c in self._path_completer.get_completions(document, complete_event):
                yield c

        # If we are inside a string, Don't do Jedi completion.
        if self._path_completer_grammar.match(document.text_before_cursor):
            return

        # Do Jedi Python completions.
        if complete_event.completion_requested or self._complete_python_while_typing(document):
            script = get_jedi_script_from_document(document, self.get_locals(), self.get_globals())

            if script:
                try:
                    completions = script.completions()
                except TypeError:
                    # Issue #9: bad syntax causes completions() to fail in jedi.
                    # https://github.com/jonathanslenders/python-prompt-toolkit/issues/9
                    pass
                except UnicodeDecodeError:
                    # Issue #43: UnicodeDecodeError on OpenBSD
                    # https://github.com/jonathanslenders/python-prompt-toolkit/issues/43
                    pass
                except AttributeError:
                    # Jedi issue #513: https://github.com/davidhalter/jedi/issues/513
                    pass
                except ValueError:
                    # Jedi issue: "ValueError: invalid \x escape"
                    pass
                except KeyError:
                    # Jedi issue: "KeyError: u'a_lambda'."
                    # https://github.com/jonathanslenders/ptpython/issues/89
                    pass
                except IOError:
                    # Jedi issue: "IOError: No such file or directory."
                    # https://github.com/jonathanslenders/ptpython/issues/71
                    pass
                except AssertionError:
                    # In jedi.parser.__init__.py: 227, in remove_last_newline,
                    # the assertion "newline.value.endswith('\n')" can fail.
                    pass
                except SystemError:
                    # File "jedi/api/helpers.py", line 140, in get_stack_at_position
                    # raise SystemError("This really shouldn't happen. There's a bug in Jedi.")
                    pass
                else:
                    for c in completions:
                        yield Completion(c.name_with_symbols, len(c.complete) - len(c.name_with_symbols),
                                         display=c.name_with_symbols)



from typing import Any

from .message import Message
from .descriptor_pb2 import (
    EnumOptions,
    EnumValueOptions,
    FieldOptions,
    FileOptions,
    MessageOptions,
    MethodOptions,
    OneofOptions,
    ServiceOptions,
)

class Error(Exception): ...
class TypeTransformationError(Error): ...

class DescriptorMetaclass(type):
    def __instancecheck__(self, obj): ...

class DescriptorBase(metaclass=DescriptorMetaclass):
    has_options: Any
    def __init__(self, options, options_class_name) -> None: ...
    def GetOptions(self): ...

class _NestedDescriptorBase(DescriptorBase):
    name: Any
    full_name: Any
    file: Any
    containing_type: Any
    def __init__(self, options, options_class_name, name, full_name, file, containing_type, serialized_start=..., serialized_end=...) -> None: ...
    def GetTopLevelContainingType(self): ...
    def CopyToProto(self, proto): ...

class Descriptor(_NestedDescriptorBase):
    def __new__(cls, name, full_name, filename, containing_type, fields, nested_types, enum_types, extensions, options=..., is_extendable=..., extension_ranges=..., oneofs=..., file=..., serialized_start=..., serialized_end=..., syntax=...): ...
    fields: Any
    fields_by_number: Any
    fields_by_name: Any
    nested_types: Any
    nested_types_by_name: Any
    enum_types: Any
    enum_types_by_name: Any
    enum_values_by_name: Any
    extensions: Any
    extensions_by_name: Any
    is_extendable: Any
    extension_ranges: Any
    oneofs: Any
    oneofs_by_name: Any
    syntax: Any
    def __init__(self, name, full_name, filename, containing_type, fields, nested_types, enum_types, extensions, options=..., is_extendable=..., extension_ranges=..., oneofs=..., file=..., serialized_start=..., serialized_end=..., syntax=...) -> None: ...
    def EnumValueName(self, enum, value): ...
    def CopyToProto(self, proto): ...
    def GetOptions(self) -> MessageOptions: ...

class FieldDescriptor(DescriptorBase):
    TYPE_DOUBLE: Any
    TYPE_FLOAT: Any
    TYPE_INT64: Any
    TYPE_UINT64: Any
    TYPE_INT32: Any
    TYPE_FIXED64: Any
    TYPE_FIXED32: Any
    TYPE_BOOL: Any
    TYPE_STRING: Any
    TYPE_GROUP: Any
    TYPE_MESSAGE: Any
    TYPE_BYTES: Any
    TYPE_UINT32: Any
    TYPE_ENUM: Any
    TYPE_SFIXED32: Any
    TYPE_SFIXED64: Any
    TYPE_SINT32: Any
    TYPE_SINT64: Any
    MAX_TYPE: Any
    CPPTYPE_INT32: Any
    CPPTYPE_INT64: Any
    CPPTYPE_UINT32: Any
    CPPTYPE_UINT64: Any
    CPPTYPE_DOUBLE: Any
    CPPTYPE_FLOAT: Any
    CPPTYPE_BOOL: Any
    CPPTYPE_ENUM: Any
    CPPTYPE_STRING: Any
    CPPTYPE_MESSAGE: Any
    MAX_CPPTYPE: Any
    LABEL_OPTIONAL: Any
    LABEL_REQUIRED: Any
    LABEL_REPEATED: Any
    MAX_LABEL: Any
    MAX_FIELD_NUMBER: Any
    FIRST_RESERVED_FIELD_NUMBER: Any
    LAST_RESERVED_FIELD_NUMBER: Any
    def __new__(cls, name, full_name, index, number, type, cpp_type, label, default_value, message_type, enum_type, containing_type, is_extension, extension_scope, options=..., file=..., has_default_value=..., containing_oneof=...): ...
    name: Any
    full_name: Any
    index: Any
    number: Any
    type: Any
    cpp_type: Any
    label: Any
    has_default_value: Any
    default_value: Any
    containing_type: Any
    message_type: Any
    enum_type: Any
    is_extension: Any
    extension_scope: Any
    containing_oneof: Any
    def __init__(self, name, full_name, index, number, type, cpp_type, label, default_value, message_type, enum_type, containing_type, is_extension, extension_scope, options=..., file=..., has_default_value=..., containing_oneof=...) -> None: ...
    @staticmethod
    def ProtoTypeToCppProtoType(proto_type): ...
    def GetOptions(self) -> FieldOptions: ...

class EnumDescriptor(_NestedDescriptorBase):
    def __new__(cls, name, full_name, filename, values, containing_type=..., options=..., file=..., serialized_start=..., serialized_end=...): ...
    values: Any
    values_by_name: Any
    values_by_number: Any
    def __init__(self, name, full_name, filename, values, containing_type=..., options=..., file=..., serialized_start=..., serialized_end=...) -> None: ...
    def CopyToProto(self, proto): ...
    def GetOptions(self) -> EnumOptions: ...

class EnumValueDescriptor(DescriptorBase):
    def __new__(cls, name, index, number, type=..., options=...): ...
    name: Any
    index: Any
    number: Any
    type: Any
    def __init__(self, name, index, number, type=..., options=...) -> None: ...
    def GetOptions(self) -> EnumValueOptions: ...

class OneofDescriptor:
    def __new__(cls, name, full_name, index, containing_type, fields): ...
    name: Any
    full_name: Any
    index: Any
    containing_type: Any
    fields: Any
    def __init__(self, name, full_name, index, containing_type, fields) -> None: ...
    def GetOptions(self) -> OneofOptions: ...

class ServiceDescriptor(_NestedDescriptorBase):
    index: Any
    methods: Any
    methods_by_name: Any
    def __init__(self, name, full_name, index, methods, options=..., file=..., serialized_start=..., serialized_end=...) -> None: ...
    def FindMethodByName(self, name): ...
    def CopyToProto(self, proto): ...
    def GetOptions(self) -> ServiceOptions: ...

class MethodDescriptor(DescriptorBase):
    name: Any
    full_name: Any
    index: Any
    containing_service: Any
    input_type: Any
    output_type: Any
    def __init__(self, name, full_name, index, containing_service, input_type, output_type, options=...) -> None: ...
    def GetOptions(self) -> MethodOptions: ...

class FileDescriptor(DescriptorBase):
    def __new__(cls, name, package, options=..., serialized_pb=..., dependencies=..., public_dependencies=..., syntax=..., pool=...): ...
    _options: Any
    pool: Any
    message_types_by_name: Any
    name: Any
    package: Any
    syntax: Any
    serialized_pb: Any
    enum_types_by_name: Any
    extensions_by_name: Any
    services_by_name: Any
    dependencies: Any
    public_dependencies: Any
    def __init__(self, name, package, options=..., serialized_pb=..., dependencies=..., public_dependencies=..., syntax=..., pool=...) -> None: ...
    def CopyToProto(self, proto): ...
    def GetOptions(self) -> FileOptions: ...

def MakeDescriptor(desc_proto, package=..., build_file_if_cpp=..., syntax=...): ...
def _ParseOptions(message: Message, string: bytes) -> Message: ...



from typing import Any
from collections import MutableMapping

class RLock:
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, traceback): ...

class RecentlyUsedContainer(MutableMapping):
    ContainerCls: Any
    dispose_func: Any
    lock: Any
    def __init__(self, maxsize=..., dispose_func=...) -> None: ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value): ...
    def __delitem__(self, key): ...
    def __len__(self): ...
    def __iter__(self): ...
    def clear(self): ...
    def keys(self): ...

class HTTPHeaderDict(dict):
    def __init__(self, headers=..., **kwargs) -> None: ...
    def __setitem__(self, key, val): ...
    def __getitem__(self, key): ...
    def __delitem__(self, key): ...
    def __contains__(self, key): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    values: Any
    get: Any
    update: Any
    iterkeys: Any
    itervalues: Any
    def pop(self, key, default=...): ...
    def discard(self, key): ...
    def add(self, key, val): ...
    def extend(self, *args, **kwargs): ...
    def getlist(self, key): ...
    getheaders: Any
    getallmatchingheaders: Any
    iget: Any
    def copy(self): ...
    def iteritems(self): ...
    def itermerged(self): ...
    def items(self): ...
    @classmethod
    def from_httplib(cls, message, duplicates=...): ...



# Stubs for email.mime.multipart (Python 3.4)

from typing import Optional, Sequence, Tuple, Union
from email.message import Message
from email.mime.base import MIMEBase

_ParamsType = Union[str, None, Tuple[str, Optional[str], str]]

class MIMEMultipart(MIMEBase):
    def __init__(self, _subtype: str = ..., boundary: Optional[str] = ...,
                 _subparts: Optional[Sequence[Message]] = ...,
                 **_params: _ParamsType) -> None: ...
